import numpy as np

import xarray as xr
from arpes.endstations import (FITSEndstation, HemisphericalEndstation,
                               SynchrotronEndstation)

__all__ = ('MAESTROMicroARPESEndstation','MAESTRONanoARPESEndstation')


class MAESTROARPESEndstationBase(SynchrotronEndstation, HemisphericalEndstation, FITSEndstation):
    """
    The MERLIN ARPES Endstation at the Advanced Light Source
    """

    PRINCIPAL_NAME = None # skip me
    ALIASES = None # skip me
    ANALYZER_INFORMATION = None

    def load(self, scan_desc: dict = None, **kwargs):
        # in the future, can use a regex in order to handle the case where we postfix coordinates
        # for multiple spectra

        scan = super().load(scan_desc, **kwargs)

        coord_names = scan.coords.keys()
        will_rename = {}
        for coord_name in coord_names:
            if coord_name in self.RENAME_KEYS:
                will_rename[coord_name] = self.RENAME_KEYS.get(coord_name)

        for k, v in will_rename.items():
            if v in scan.coords:
                del scan.coords[v]

        renamed = scan.rename(will_rename)

        if 'scan_x' in renamed.coords:
            for d in renamed.data_vars:
                if 'spectrum' in d:
                    renamed[d].values = np.flip(renamed[d].values, axis=renamed[d].dims.index('scan_x'))

        return renamed

    def fix_prebinned_coordinates(self):
        pass

    def postprocess_final(self, data: xr.Dataset, scan_desc: dict=None):
        ls = [data] + data.S.spectra
        for l in ls:
            l.attrs.update(self.ANALYZER_INFORMATION)

            if 'GRATING' in l.attrs:
                l.attrs['grating_lines_per_mm'] = {
                    'G201b': 600,
                }.get(l.attrs['GRATING'])

        data = super().postprocess_final(data, scan_desc)

        return data


class MAESTROMicroARPESEndstation(MAESTROARPESEndstationBase):
    """
    Implements data loading at the microARPES endstation of ALS's MAESTRO.
    """
    PRINCIPAL_NAME = 'ALS-BL7'
    ALIASES = ['BL7', 'BL7.0.2', 'ALS-BL7.0.2', 'MAESTRO']

    ANALYZER_INFORMATION = {
        'analyzer': 'R4000',
        'analyzer_name': 'Scienta R4000',
        'parallel_deflectors': False,
        'perpendicular_deflectors': True,
        'analyzer_radius': None,
        'analyzer_type': 'hemispherical',
    }

    RENAME_KEYS = {
        'LMOTOR0': 'x',
        'LMOTOR1': 'y',
        'LMOTOR2': 'z',
        'Scan X': 'scan_x',
        'Scan Y': 'scan_y',
        'Scan Z': 'scan_z',
        'LMOTOR3': 'theta',
        'LMOTOR4': 'beta',
        'LMOTOR5': 'chi',
        'LMOTOR6': 'alpha',
        'LMOTOR9': 'psi',
        'mono_eV': 'hv',
        'SF_HV': 'hv',
        'SS_HV': 'hv',
        'Slit Defl': 'psi',

        # probably need something like an attribute list for extraction
        'SFRGN0': 'fixed_region_name',
        'SFE_0': 'daq_center_energy',
        'SFLNM0': 'lens_mode_name',
        'SFPE_0': 'pass_energy',
        'UNDHARM': 'undulator_harmonic',
        'RINGCURR': 'beam_current',
        'SFFR_0': 'frames_per_slice',

        'SFBA_0': 'phi_prebinning',
        'SFBE0': 'eV_prebinning',
        'LWLVNM': 'daq_type',
    }

    RENAME_COORDS = {
        'X': 'x',
        'Y': 'y',
        'Z': 'z',
    }

    ATTR_TRANSFORMS = {
        'START_T': lambda l: {'time': ' '.join(l.split(' ')[1:]).lower(),
                              'date': l.split(' ')[0]},
        'SF_SLITN': lambda l: {'slit_number': int(l.split(' ')[0]),
                               'slit_shape': l.split(' ')[-1].lower(),
                               'slit_width': float(l.split(' ')[2])},
    }

    MERGE_ATTRS = {
        'mcp_voltage': None,
        'repetition_rate': 5e8,
        'undulator_type': 'elliptically_polarized_undulator',
        'undulator_gap': None,
        'undulator_z': None,
        'undulator_polarization': None,
    }


class MAESTRONanoARPESEndstation(MAESTROARPESEndstationBase):
    """
    Implements data loading at the nanoARPES endstation of ALS's MAESTRO.
    """
    PRINCIPAL_NAME = 'ALS-BL7-nano'
    ALIASES = ['BL7-nano', 'BL7.0.2-nano', 'ALS-BL7.0.2-nano', 'MAESTRO-nano']

    ENSURE_COORDS_EXIST = [
        'long_x', 'long_y', 'long_z', 'short_x', 'short_y', 'short_z',
        'theta', 'beta', 'chi', 'hv', 'alpha', 'psi',
        'physical_long_x', 'physical_long_y', 'physical_long_z',
    ]

    ANALYZER_INFORMATION = {
        'analyzer': 'DA-30',
        'analyzer_name': 'Scienta DA-30',
        'parallel_deflectors': False,
        'perpendicular_deflectors': False,
        'analyzer_radius': None,
        'analyzer_type': 'hemispherical',
    }

    RENAME_KEYS = {
        'LMOTOR0': 'long_x',
        'LMOTOR1': 'long_z',
        'LMOTOR2': 'long_y',

        'PMOTOR0': 'physical_long_x',
        'PMOTOR1': 'physical_long_z',
        'PMOTOR2': 'physical_long_y',

        'LMOTOR3': 'chi',
        'LMOTOR4': 'theta',

        'LMOTOR5': 'order_sorting_aperature_x',
        'LMOTOR6': 'order_sorting_aperature_z',
        'LMOTOR7': 'order_sorting_aperature_y',

        'LMOTOR8': 'optics_insertion',
        'LMOTOR9': 'optics_select',

        'LMOTOR10': 'short_x',  # these are the scan stages
        'LMOTOR11': 'short_z',
        'LMOTOR12': 'short_y',

        'LMOTOR13': 'psi',  # Chamber Theta
        'LMOTOR14': 'alpha',

        'LMOTOR15': 'optics_pitch',
        'LMOTOR16': 'optics_yaw',

        'LMOTOR17': 'aperature_x',
        'LMOTOR18': 'aperature_y',

        'LMOTOR19': 'sample_float',

        'MONO_E': 'hv',
        'E_RESOLU': 'probe_linewidth',
        'EPU_E': 'undulator_hv',
        'EPU_POL': 'probe_polarization',
        'EPU_Z': 'undulator_z',
        'EPU_GAP': 'undulator_gap',

        'UNDHARM': 'undulator_harmonic',
        'RINGCURR': 'beam_current',

        'SF_HV': 'hv',
        'SS_HV': 'hv',

        # probably need something like an attribute list for extraction
        'SFRGN0': 'fixed_region_name',
        'SFE_0': 'daq_center_energy',
        'SFLNM0': 'lens_mode_name',
        'SFPE_0': 'pass_energy',

        'SFFR_0': 'frames_per_slice',

        'SFBA_0': 'phi_prebinning',
        'SFBE0': 'eV_prebinning',
        'LWLVNM': 'daq_type',
    }

    RENAME_COORDS = {
        'X': 'long_x',
        'Y': 'long_y',
        'Z': 'long_z',
        'Scan X': 'short_x',
        'Scan Y': 'short_y',
        'Scan Z': 'short_z',
        'Sample X': 'long_x',
        'Sample Y': 'long_y',
        'Sample Z': 'long_z',
        'Optics Stage': 'optics_insertion',
        'Pitch': 'pitch',
    }

    ATTR_TRANSFORMS = {
        'START_T': lambda l: {'time': ' '.join(l.split(' ')[1:]).lower(),
                              'date': l.split(' ')[0]},
        'SF_SLITN': lambda l: {'slit_number': int(l.split(' ')[0]),
                               'slit_shape': l.split(' ')[-1].lower(),
                               'slit_width': float(l.split(' ')[2])},
    }

    MERGE_ATTRS = {
        'mcp_voltage': None,
        'beta': 0,
        'repetition_rate': 5e8,
        'undulator_type': 'elliptically_polarized_undulator',
        'undulator_gap': None,
        'undulator_z': None,
        'undulator_polarization': None,
    }

    @staticmethod
    def update_hierarchical_coordinates(data: xr.Dataset):
        """
        Nano-ARPES endstations often have two sets of spatial coordinates, a long-range piezo inertia
        or stepper stage, sometimes outside vacuum, and a fast, high resolution piezo scan stage that may or may not
        be based on piezo inertia ("slip-stick") type actuators.

        Additionally, any spatially imaging experiments like PEEM or the transmission operating mode
        of hemispherical analyzers have two spatial coordinates, the one on the manipulator and the
        imaged axis. In these cases, this imaged axis will always be treated in the same role as the high-resolution
        motion axis of a nano-ARPES system.

        Working in two coordinate systems is frustrating, and it makes comparing data cumbersome. In PyARPES
        x,y,z is always the total inferrable coordinate value, i.e. (+/- long range +/- high resolution)
        as appropriate. You can still access the underlying coordinates in this case as
        `long_{dim}` and `short_{dim}`.

        :param data:
        :return:
        """

        ls = [data] + data.S.spectra

        for d_name in ['x', 'y', 'z']:
            short, long = 'short_{}'.format(d_name), 'long_{}'.format(d_name)
            phys = 'physical_long_{}'.format(d_name)

            def lookup(name):
                coordinate = data.S.lookup_coord(name)
                try:
                    return coordinate.values
                except AttributeError:
                    return coordinate

            c_short, c_long = lookup(short), lookup(long)

            data.coords[phys] = -data.coords[phys]

            scan_coord_name = None
            if isinstance(c_short, np.ndarray):
                scan_coord_name = short
            elif isinstance(c_long, np.ndarray):
                scan_coord_name = long

            if scan_coord_name:
                data.coords[scan_coord_name].values = -c_short - c_long
                data = data.rename(dict([[scan_coord_name, d_name]]))
                data.coords[short if scan_coord_name == short else long] = -c_short if scan_coord_name == short else -c_long
            else:
                data.coords[d_name] = -c_short - c_long

        return data

    @staticmethod
    def unwind_serptentine(data: xr.Dataset) -> xr.Dataset:
        """
        MAESTRO supports a serpentine (think snake the computer game) scan mode to minimize the motion time for
        coarsely locating samples. Unfortunately, the DAQ just dumps the raw data, so we have to unwind it ourselves.
        :param data:
        :return:
        """

        spectra = data.S.spectra
        for spectrum in spectra:
            # serpentine axis always seems to be the first one, thankfully
            old_values = spectrum.values.copy()

            mask = np.mod(np.array(range(len(spectrum.coords[spectrum.dims[0]]))), 2) == 1
            old_values[mask] = np.roll(old_values[mask,::-1], 2, axis=1)

            spectrum.values = old_values

        return data

    def postprocess_final(self, data: xr.Dataset, scan_desc: dict=None):
        data = data.rename({k: v for k, v in self.RENAME_COORDS.items() if k in data.coords.keys()})
        data = super().postprocess_final(data, scan_desc)

        # microns to mm
        ls = [data] + data.S.spectra
        for c in ['short_x', 'short_y', 'short_z',
                  'physical_long_x', 'physical_long_y', 'physical_long_z']:
            for l in ls:
                if c in l.attrs:
                    l.attrs[c] = l.attrs[c] / 1000

                l.coords[c] = l.coords[c] / 1000

        data = MAESTRONanoARPESEndstation.update_hierarchical_coordinates(data)
        if data.attrs['daq_type'] == 'MotorSerpentine':
            data = MAESTRONanoARPESEndstation.unwind_serptentine(data)

        # we return new data from update_hierarchical, so we need to refresh
        # the definition of ls
        ls = [data] + data.S.spectra
        for l in ls:
            l.coords['alpha'] = np.pi / 2
            l.attrs['alpha'] = np.pi / 2

            l.attrs['phi_offset'] = 0.4

            l.coords['phi'] = l.coords['phi'] / 2

        for deg_to_rad_coord in {'theta', 'psi', 'beta', }:
            for l in ls:
                l.coords[deg_to_rad_coord] = l.coords[deg_to_rad_coord] * np.pi / 180
                if deg_to_rad_coord in l.attrs:
                    l.attrs[deg_to_rad_coord] = l.attrs[deg_to_rad_coord] * np.pi / 180

        return data



"""
arpes.moire includes some tools for analyzing moirés and data on moiré heterostructures in particular.

All of the moirés discussed here are on hexagonal crystal systems.
"""

import numpy as np
import matplotlib.pyplot as plt

from arpes.utilities.bz import hex_cell_2d
from arpes.plotting.bz import bz_plot, Translation, Rotation


def calculate_bz_vertices_from_direct_cell(cell):
    from ase.dft.bz import bz_vertices

    if len(cell) > 2:
        assert all(abs(cell[2][0:2]) < 1e-6) and all(abs(cell.T[2][0:2]) < 1e-6)
    else:
        cell = [list(c) + [0] for c in cell] + [[0, 0, 1]]

    icell = np.linalg.inv(cell).T
    try:
        bz1 = bz_vertices(icell[:3, :3], dim=2)
    except TypeError:
        bz1 = bz_vertices(icell[:3, :3])

    return bz1


def angle_between_vectors(a, b):
    return np.arccos(np.dot(a,b) / (np.linalg.norm(a) * np.linalg.norm(b)))


def calc_commensurate_moire_cell(underlayer_a, overlayer_a, relative_angle=0, swap_angle=False):
    """
    Calculates nearly commensurate moire unit cells for two hexagonal lattices
    :return:
    """
    from ase.dft.kpoints import get_special_points
    from ase.dft.bz import bz_vertices

    underlayer_direct = hex_cell_2d(a=underlayer_a)
    overlayer_direct = hex_cell_2d(a=overlayer_a)

    underlayer_direct = [list(c) + [0] for c in underlayer_direct] + [[0, 0, 1]]
    overlayer_direct = [list(c) + [0] for c in overlayer_direct] + [[0, 0, 1]]

    underlayer_icell = np.linalg.inv(underlayer_direct).T
    overlayer_icell = np.linalg.inv(overlayer_direct).T

    underlayer_k = np.dot(underlayer_icell.T, get_special_points(underlayer_direct)['K'])
    overlayer_k = Rotation.from_rotvec([0, 0, relative_angle]).apply(
        np.dot(overlayer_icell.T, get_special_points(overlayer_direct)['K']))

    moire_k = (underlayer_k - overlayer_k)
    moire_a = underlayer_a * (np.linalg.norm(underlayer_k) / np.linalg.norm(moire_k))
    moire_angle = angle_between_vectors(underlayer_k, moire_k)

    if swap_angle:
        moire_angle = -moire_angle

    moire_cell = hex_cell_2d(moire_a)
    moire_cell = [list(c) + [0] for c in moire_cell] + [[0, 0, 1]]
    moire_cell = Rotation.from_rotvec([0, 0, moire_angle]).apply(moire_cell)
    moire_icell = np.linalg.inv(moire_cell).T

    moire_bz_points = bz_vertices(moire_icell)
    moire_bz_points = moire_bz_points[[len(p[0]) for p in moire_bz_points].index(6)][0]

    return {
        'k_points': (underlayer_k, overlayer_k, moire_k),
        'moire_a': moire_a,
        'moire_k': moire_k,
        'moire_cell': moire_cell,
        'moire_icell': moire_icell,
        'moire_bz_points': moire_bz_points,
        'moire_bz_angle': moire_angle,
    }


def plot_simple_moire_unit_cell(underlayer_a, overlayer_a, relative_angle, ax=None, offset=True,
                                swap_angle=False):
    """
    Plots a digram of a moiré unit cell.
    :return:
    """

    if ax is None:
        _, ax = plt.subplots()

    bz_plot(cell=hex_cell_2d(a=underlayer_a), linewidth=1, ax=ax, paths=[], hide_ax=False, set_equal_aspect=False)
    bz_plot(cell=hex_cell_2d(a=overlayer_a), linewidth=1, ax=ax, paths=[],
            transformations=[Rotation.from_rotvec([0, 0, relative_angle])], hide_ax=False, set_equal_aspect=False)

    moire_info = calc_commensurate_moire_cell(underlayer_a, overlayer_a, relative_angle, swap_angle=swap_angle)
    moire_k = moire_info['moire_k']

    if offset:
        k_offset = Rotation.from_rotvec([0, 0, 120 * np.pi / 180]).apply(moire_k)
    else:
        k_offset = 0

    bz_plot(cell=hex_cell_2d(a=moire_info['moire_a']), linewidth=1, ax=ax, paths=[],
            transformations=[Rotation.from_rotvec([0, 0, -moire_info['moire_bz_angle']]),
                             Translation(moire_info['k_points'][0] + k_offset)],
            hide_ax=True, set_equal_aspect=True)




import pytest
import warnings

import arpes.config


def load_test_scan(dataset_name, file_number):
    from arpes.utilities.dataset import default_dataset
    from arpes.io import direct_load

    with warnings.catch_warnings():

        warnings.simplefilter("ignore")

        df = default_dataset(match=dataset_name, write=False)
        return direct_load(file_number, df=df)



import warnings

import numpy as np
import xarray as xr

__all__ = ('imread_to_xarray', 'imread',)


def imread(str_or_path):
    """
    A wrapper around `opencv.imread` and `imageio.imread` that

    1. Falls back to the first available option on the system
    2. Normalizes OpenCV images to RGB format
    3. Removes the alpha channel from imageio.imread data

    :param str_or_path: pathlib.Path or str containing the image to be read
    :return:
    """
    try:
        import cv2
        using_cv2 = True
    except ImportError:
        try:
            import imageio

            using_cv2 = False
        except ImportError as e:
            warnings.warn('You need OpenCV or imageio in order to read images in PyARPES.')
            raise e

    if using_cv2:
        arr = cv2.imread(str(str_or_path))
        return np.stack([arr[:, :, 2], arr[:, :, 1], arr[:, :, 0]], axis=-1)
    else:
        arr = imageio.imread(str(str_or_path))
        return arr[:,:,:3]


def imread_to_xarray(str_or_path):
    """
    Like `imread`, except that this function wraps the result into a xr.DataArray
    that has x (pixel), y (pixel), and color (['R', 'G', 'B']) dimensions.

    :param str_or_path:
    :return:
    """
    raw_arr = imread(str_or_path)
    sx, sy, _ = raw_arr.shape

    return xr.DataArray(
        raw_arr,
        coords={
            'x': np.asarray(range(sx)),
            'y': np.asarray(range(sy)),
            'color': ['r', 'g', 'b']
        },
        dims=['x', 'y', 'color']
    )



import datetime
import json
import os
import urllib.request

from pathlib import Path
from tqdm import tqdm_notebook
from typing import List, Optional


__all__ = ('get_full_notebook_information', 'get_notebook_name', 'generate_logfile_path',
           'get_recent_logs', 'get_recent_history', 'wrap_tqdm')


def wrap_tqdm(x, interactive=True, *args, **kwargs):
    if not interactive:
        return x

    return tqdm_notebook(x, *args, **kwargs)


def get_full_notebook_information() -> Optional[dict]:
    """
    Javascriptless method to get information about the current Jupyter sessions and the one matching
    this kernel.
    :return:
    """

    try:  # Respect those that opt not to use IPython
        from notebook import notebookapp
        import ipykernel
    except ImportError:
        return None

    connection_file = os.path.basename(ipykernel.get_connection_file())
    kernel_id = connection_file.split('-', 1)[1].split('.')[0]

    servers = notebookapp.list_running_servers()
    for server in servers:
        try:
            passwordless = not server['token'] and not server['password']
            url = server['url']+'api/sessions' + ('' if passwordless else '?token={}'.format(server['token']))
            sessions = json.load(urllib.request.urlopen(url))
            for sess in sessions:
                if sess['kernel']['id'] == kernel_id:
                    return {'server': server, 'session': sess,}
        except:
            pass
    return None


def get_notebook_name() -> Optional[str]:
    """
    Gets the unqualified name of the running Jupyter notebook, if there is a Jupyter session
    not protected by password.

    As an example, if you were running a notebook called "Doping-Analysis.ipynb"
    this would return "Doping-Analysis".

    If no notebook is running for this kernel or the Jupyter session is password protected, we
    can only return None.
    :return:
    """

    jupyter_info = get_full_notebook_information()

    try:
        return jupyter_info['session']['notebook']['name'].split('.')[0]
    except (KeyError, TypeError):
        return None


def generate_logfile_path() -> Path:
    """
    Generates a time and date qualified path for the notebook log file.
    :return:
    """

    base_name = get_notebook_name() or 'unnamed'
    full_name = '{}_{}_{}.log'.format(
        base_name,
        datetime.date.today().isoformat(),
        datetime.datetime.now().time().isoformat().split('.')[0].replace(':', '-')
    )
    return Path('logs') / full_name


def get_recent_history(n_items=10) -> List[str]:
    try:
        import IPython
        ipython = IPython.get_ipython()

        return [l[-1] for l in list(ipython.history_manager.get_tail(n=n_items, include_latest=True))]
    except (ImportError, AttributeError):
        return ['No accessible history.']


def get_recent_logs(n_bytes=1000) -> List[str]:
    import arpes.config
    try:
        import IPython
        ipython = IPython.get_ipython()
        if arpes.config.CONFIG['LOGGING_STARTED']:
            logging_file = arpes.config.CONFIG['LOGGING_FILE']

            print(logging_file)
            with open(logging_file, 'rb') as file:
                try:
                    file.seek(-n_bytes, os.SEEK_END)
                except OSError:
                    file.seek(0)

                lines = file.readlines()

            # ensure we get the most recent information
            final_cell = ipython.history_manager.get_tail(n=1, include_latest=True)[0][-1]
            return [l.decode() for l in lines] + [final_cell]

    except (ImportError, AttributeError):
        pass

    return ['No logging available. Logging is only available inside Jupyter.']



import sys
import warnings

import arpes.config

import numpy as np
import xarray as xr

from PyQt5 import QtGui, QtCore, QtWidgets
from matplotlib.backends.backend_qt5agg import FigureCanvas
from matplotlib.figure import Figure

from arpes.utilities.conversion import convert_coordinates
from arpes.utilities.bz_spec import SURFACE_ZONE_DEFINITIONS
from arpes.utilities.image import imread_to_xarray
from arpes.plotting.qt_tool.utils import (
    pretty_key_event, PRETTY_KEYS, KeyBinding, hlayout, vlayout, layout, tabs, combobox
)
from arpes.plotting.utils import imshow_arr
from arpes.plotting.qt_tool.excepthook import patched_excepthook

from .CoordinateOffsetWidget import CoordinateOffsetWidget


class BZToolWindow(QtGui.QMainWindow, QtCore.QObject):
    def __init__(self, *args, **kwargs):
        super().__init__()
        self.tool = None
        self._old_excepthook = sys.excepthook
        sys.excepthook = patched_excepthook

        self._keyBindings = [
            KeyBinding('Close', [QtCore.Qt.Key_Escape], self.do_close),
        ]

        QtGui.QGuiApplication.installEventFilter(self, self)

    def close(self):
        sys.excepthook = self._old_excepthook
        super().close()

    def do_close(self, event):
        self.close()

    def window_print(self, *args, **kwargs):
        print(*args, **kwargs)

    def eventFilter(self, source, event):
        special_keys = [QtCore.Qt.Key_Down, QtCore.Qt.Key_Up, QtCore.Qt.Key_Left, QtCore.Qt.Key_Right]

        if event.type() in [QtCore.QEvent.KeyPress, QtCore.QEvent.ShortcutOverride]:
            if event.type() != QtCore.QEvent.ShortcutOverride or event.key() in special_keys:
                self.handleKeyPressEvent(event)

        return super().eventFilter(source, event)

    def handleKeyPressEvent(self, event):
        handled = False
        for binding in self._keyBindings:
            for combination in binding.chord:
                # only detect single keypresses for now
                if combination == event.key():
                    handled = True
                    binding.handler(event)

        if not handled:
            if arpes.config.SETTINGS.get('DEBUG', False):
                print(event.key())


class BZTool:
    """
    Implements a Brillouin zone explorer showing the region of momentum
    probed by ARPES.
    """

    def __init__(self):
        self.settings = None
        self.context = {}

        self.content_layout = None
        self.main_layout = None
        self.views = {}
        self.reactive_views = []
        self.current_material = None
        self.cut_line = None

        self.canvas = None
        self.ax = None

    def configure_main_widget(self):
        self.canvas = FigureCanvas(Figure(figsize=(8,8)))
        self.ax = self.canvas.figure.subplots()
        self.content_layout.addWidget(self.canvas, 0, 0)

    def on_change_material(self, value):
        self.current_material = SURFACE_ZONE_DEFINITIONS[value]

        if 'bz_points' in self.current_material:
            bz_points = self.current_material['bz_points']
            try:
                bz_points = bz_points()
            except TypeError:
                pass

            x, y = np.concatenate([bz_points, bz_points[:1]]).T
            self.ax.clear()
            self.ax.plot(x, y, linewidth=2)
            self.ax.set_aspect(1)
        else:
            assert 'image' in self.current_material
            image = self.current_material['image']
            image_waypoints = self.current_material['image_waypoints']
            if not image_waypoints[0]:
                warnings.warn('Missing waypoints for material: {}'.format(self.current_material['name']))
                image_waypoints = [
                    [0, 1, 0, 1],
                    [1, 0, 1, 0],
                ]

            self.image_waypoints = image_waypoints
            self.image_data = imread_to_xarray(image)

            # the way we are currently using waypoints assumes that the data has orthogonal
            # axes and they are not rotated. This might not be appropriate in all circumstances,
            # but it is how we are assuming things for now

            way_p, way_pp = image_waypoints
            way_p_x, way_p_y, way_p_kx, way_p_ky = way_p
            way_pp_x, way_pp_y, way_pp_kx, way_pp_ky = way_pp

            dx, dy, dkx, dky = ((way_pp_x - way_p_x), (way_pp_y - way_p_y),
                                (way_pp_kx - way_p_kx), (way_pp_ky - way_p_ky))

            self.image_data.coords['y'].values = (self.image_data.coords['y'].values - way_p_x) \
                                                 * (dkx / dx) + way_p_kx
            self.image_data.coords['x'].values = (self.image_data.coords['x'].values - way_p_y) \
                                                 * (-dky / dy) - way_p_ky

            self.ax.clear()
            self.image_data.values = self.image_data.values[::-1,:]
            imshow_arr(self.image_data, ax=self.ax, origin='lower')
            self.ax.scatter([0], [0], s=100, color='green')
            self.ax.set_aspect(1)

        self.ax.figure.canvas.draw()
        self.update_cut()

    @property
    def coordinates(self):
        return {k: np.pi * v.value() / 180 for k, v in self.coordinate_widgets.items()}

    def update_cut(self, *args):
        coords = {
            'phi': np.linspace(-15., 15., 51) * np.pi / 180,
            'hv': 5.93, # FOR NOW
            'eV': 0,
        }

        coords_from_widgets = self.coordinates
        coords.update({k: v for k, v in coords_from_widgets.items() if k not in {'phi', }})

        cut = xr.Dataset(data_vars={}, coords=coords, attrs={
            'work_function': self.current_material.get('work_function', 4.2) or 4.2,
            'inner_potential': self.current_material.get('inner_potential', 10) or 10,
            'phi_offset': coords_from_widgets['phi'],
        })

        kcut = convert_coordinates(cut)

        try:
            self.ax.lines.remove(self.cut_line[0])
        except Exception:
            pass

        self.cut_line = self.ax.plot(kcut.kx.values, kcut.ky.values, self.current_material.get('cut_color', 'green'), linewidth=3)
        self.ax.figure.canvas.draw()

    def construct_coordinate_info_tab(self):
        needed_coordinates = ['phi', 'psi', 'alpha', 'theta', 'beta', 'chi']
        inner_items = [CoordinateOffsetWidget(coordinate_name=coordinate, root=self)
                       for coordinate in needed_coordinates]
        self.coordinate_widgets = dict(zip(needed_coordinates, inner_items))

        for widget in inner_items:
            pass
        return hlayout(*inner_items)

    def construct_sample_info_tab(self):
        material_choice = combobox(
            name='Material Specification',
            items=sorted(SURFACE_ZONE_DEFINITIONS.keys())
        )
        self.material_choice_widget = material_choice
        self.material_choice_widget.currentTextChanged.connect(self.on_change_material)
        inner_items = [material_choice]
        return hlayout(*inner_items)

    def construct_detector_info_tab(self):
        inner_items = []
        return hlayout(*inner_items)

    def construct_general_settings_tab(self):
        inner_items = []
        return hlayout(*inner_items)

    def add_coordinate_control_widgets(self):
        self.coordinate_info_tab = self.construct_coordinate_info_tab()
        self.sample_info_tab = self.construct_sample_info_tab()
        self.detector_info_tab = self.construct_detector_info_tab()
        self.general_settings_tab = self.construct_general_settings_tab()

        self.tabs = tabs(
            ['Coordinates', self.coordinate_info_tab],
            ['Sample Info', self.sample_info_tab],
            ['Detector Type and Settings', self.detector_info_tab],
            #['Beamline Info', self.beamline_info],
            ['General Settings', self.general_settings_tab],
        )

        self.tabs.setFixedHeight(250)
        self.main_layout.addLayout(self.content_layout, 0, 0)
        self.main_layout.addWidget(self.tabs, 1, 0)

    def start(self):
        app = QtGui.QApplication([])

        win = BZToolWindow()
        win.resize(1100 - 250,1100)
        win.setWindowTitle('Brillouin Zone Tool')
        cw = QtGui.QWidget()
        win.setCentralWidget(cw)
        self.window = win
        self.window.tool = self

        self.content_layout = QtGui.QGridLayout()
        self.main_layout = QtGui.QGridLayout()

        cw.setLayout(self.main_layout)

        self.configure_main_widget()
        self.add_coordinate_control_widgets()

        self.material_choice_widget.setCurrentText('Td-WTe2')

        self.window.show()

        QtGui.QApplication.instance().exec()


def bz_tool():
    tool = BZTool()
    tool.start()